#include <Arduino.h>
#include <ETH.h>
#include <IPAddress.h>
#include <SPI.h>
#include <PubSubClient.h>
#include <SPIFFS.h>
#include <HTTPUpdate.h>
#include <ADS1231.h>

#define LED_SIZE 10
#define BUTTON1 35
#define ETH_NRST 14
#define LED_BLUE 17
#define WIFI_CLIENT_DEF_CONN_TIMEOUT_MS 5000
#define ARDUHAL_LOG_LEVEL 0
#define ADS_PDWN 5
#define ADS_SCLK 4
#define ADS_DRDY_DOUT 2

static bool eth_connected = false;
bool backupServer = false;
int rebounds = 0;
const int NUM_TOPIC_SUB = 7;
const int NUM_TOPIC_PUB = 2;
long lastReconnectAttempt = 0;
long lastValues = 0;
double raw = 0;
double weight = 0;
String broker_port;
String DR_ID;
String mqtt_cient;
String subscribe_topic;
String LWTtopic;

WiFiClient ethClient;
PubSubClient client(ethClient);
ADS1231 adc(ADS_SCLK,ADS_DRDY_DOUT);

struct config {
  float tare;
	double scale;
  double zero;
	int frequency;
  String otaUrl;
	IPAddress ServerIP;
	int ServerPort;
}DR;

String TopicsToSubscribe[NUM_TOPIC_SUB] = {
		"/tare", "/config/scale", "/config/zero",
		"/config/frequency", "/config/server_ip", 
    "/OTA", "/Update"};

String TopicsToPublish[NUM_TOPIC_PUB] = {
		"/weight",
		"/raw" };

void updateOTA(String ota_url){

  rebounds = 0;
  WiFiClient client;
  t_httpUpdate_return ret = httpUpdate.update(client, ota_url);

  switch (ret)
  {
  case HTTP_UPDATE_FAILED:
    Serial.printf("HTTP_UPDATE_FAILED Error (%d): %s\n", httpUpdate.getLastError(), httpUpdate.getLastErrorString().c_str());
    break;

  case HTTP_UPDATE_NO_UPDATES:
    Serial.println("HTTP_UPDATE_NO_UPDATES");
    break;

  case HTTP_UPDATE_OK:
    Serial.println("HTTP_UPDATE_OK");
    break;
  }
}

bool reconnect() {
  // Loop until we're reconnected
  Serial.println("Server to connect: " + DR.ServerIP.toString() + ":" + DR.ServerPort);
  Serial.print("Attempting MQTT connection...");

  // Attempt to connect
  if (client.connect(mqtt_cient.c_str(),LWTtopic.c_str(),2,true,"Offline")) {
    Serial.println("connected");
    for(int i = 0; i < NUM_TOPIC_SUB; i++){
      client.subscribe(TopicsToSubscribe[i].c_str());
      Serial.print("Subscribing to: ");
      Serial.println(TopicsToSubscribe[i]);
    }
    client.publish(LWTtopic.c_str(), "Online", true);
    return true;
  } else {
    Serial.print("failed, rc=");
    Serial.print(client.state());
    Serial.println(" try again in 5 seconds");
    // Wait 5 seconds before retrying
  }
  return false;
}

void readSPIFFS(){
  if (SPIFFS.exists("/MQTT.txt"))
  {
    String line;
    String broker;
    String port;
    File file_MQTT = SPIFFS.open("/MQTT.txt");
    line  = file_MQTT.readString();
    Serial.println(line);
    
    broker_port = line;
    broker = broker_port.substring(0,broker_port.indexOf(":"));
    port = broker_port.substring(broker_port.indexOf(":")+1);
    DR.ServerIP.fromString(broker);
    DR.ServerPort = port.toInt();

    Serial.println("Server IP: " + DR.ServerIP.toString());
    Serial.println("Server Port: " + String(DR.ServerPort));
    backupServer = true;

    file_MQTT.close();
  }
}

void callback(char* topic, byte* payload, unsigned int length) {

  String Stopic = String(topic);
  String Spayload;
  byte task;

  Serial.print("Message arrived [");
  Serial.print(topic);
  Serial.println("] ");

  for (int i = 0; i < length; i++)
  {
    Spayload += (char)payload[i];
  }

  if (Stopic == (DR_ID + "/tare")){task = 1;}
  else if (Stopic == (DR_ID + "/config/scale")){task = 2;}
  else if (Stopic == (DR_ID + "/config/zero")){task = 3;}
  else if (Stopic == (DR_ID + "/config/frequency")){task = 4;}
  else if (Stopic == (DR_ID + "/config/server_ip")){task = 5;}
  else if (Stopic == (DR_ID + "/OTA")){task = 6;}
  else if (Stopic == (DR_ID + "/Update")){task = 7;}
  
  switch (task){
    case 1:{
      DR.tare = Spayload.toFloat();
      break;
    }
    case 2:{
      DR.scale = Spayload.toDouble();
      break;
    }
    case 3:{
      DR.zero = Spayload.toDouble();
      break;
    }
    case 4:{
      DR.frequency = Spayload.toInt();
      break;
    }
    case 5:{
      Serial.println(Spayload);
      if(DR.ServerIP.toString() != Spayload.substring(0,Spayload.indexOf(":")) || DR.ServerPort != (Spayload.substring(Spayload.indexOf(":")+1)).toInt())
      {
        DR.ServerIP.fromString(Spayload.substring(0,Spayload.indexOf(":")));
        DR.ServerPort = (Spayload.substring(Spayload.indexOf(":")+1)).toInt();
        Serial.println("ServerIP: " + DR.ServerIP.toString());
        Serial.println("ServerPort: " + String(DR.ServerPort));

        File file_MQTT = SPIFFS.open("/MQTT.txt", "w");
        Serial.println(Spayload);
        file_MQTT.print(Spayload);
        file_MQTT.flush();
        file_MQTT.close();
        File file_MQTT1 = SPIFFS.open("/MQTT.txt");
        String line  = file_MQTT1.readString();
        Serial.println(line);
        file_MQTT1.close();

        client.disconnect();
        delay(10);
        client.setServer(DR.ServerIP, DR.ServerPort);
        delay(10);
        reconnect();
      }
      break;
    }
    case 6:{
      Serial.println(Spayload);
      DR.otaUrl = Spayload;
      break;
    }
    case 7:{
      Serial.println("Updating...");
      updateOTA(DR.otaUrl);
      break;
    }
  }
}

void fillTopics(){
  for(int i = 0; i < NUM_TOPIC_PUB; i++){
    TopicsToPublish[i] = DR_ID + TopicsToPublish[i];
  }
  for(int j = 0; j < NUM_TOPIC_SUB; j++){
    TopicsToSubscribe[j] = DR_ID + TopicsToSubscribe[j];
  }
}

void initialiseDR(){
  // Initialization
  DR.tare = 0;
  DR.zero = 0;
  DR.scale = 1;
  DR.frequency = 1000;
  DR.ServerIP = (0,0,0,0);
}

void initialiseMQTT(){
  DR_ID = ETH.macAddress();
  mqtt_cient = "DR_Client" + DR_ID;
  LWTtopic = DR_ID + "/Status";
  if(!backupServer){
    DR.ServerIP = ETH.gatewayIP();
    DR.ServerPort = 1883;
  }
  else{
    String broker;
    String port;
    Serial.println(broker_port);
    broker = broker_port.substring(0,broker_port.indexOf(":"));
    port = broker_port.substring(broker_port.indexOf(":")+1);
    DR.ServerIP.fromString(broker);
    DR.ServerPort = port.toInt();
  }
  Serial.println("DR_ID: " + DR_ID);
  Serial.println("broker_port: " + DR.ServerIP.toString() + ":" + String(DR.ServerPort));

  fillTopics();
  
  // client.setServer(DR.ServerIP, DR.ServerPort);
}

void readValue(){
  long val;
  double voltage;
  float Vref = 5.165;
  int counts = 0;
  double media = 0;
  long time1 = millis();

  for(int i = 0; i < 10; i++){
    if (adc.getValue(val)) {        //this call blocks until a sample is ready!
     
      counts += 1;
      // Serial.println("Value read: " + String(val));
      voltage = 0.5*val*Vref/(128*8388607);
      voltage = round(voltage*100000);
      voltage = voltage/100000;

      // Serial.print("Voltage value: ");
      // Serial.println(voltage,10);

      media += voltage;
      
    } else {
      Serial.println("Failed to get data");
    }
  }
  long time2 = millis();
  long time3 = time2-time1;
  Serial.print("time took: ");
  Serial.println(time3);

  media = media/counts;
  Serial.print("Voltage value: ");
  Serial.println(media,6);

  // raw = (40000*0.5/(128*0.002*8388607))*val;  //23 bits of accuracy (24th is sign) with 4.2 volt (measured) AVDD. 2^23 = 8388607 !
  raw = (40000/(Vref*0.002))*media;  //23 bits of accuracy (24th is sign) with 4.2 volt (measured) AVDD. 2^23 = 8388607 !
  Serial.print("RAW value: ");
  Serial.println(raw,6);

  weight = DR.scale*raw - DR.zero - DR.tare;
  Serial.println("Final Weight: " + String(weight));
}

void sendValue(){
  char textmessage[32];

  sprintf(textmessage, "%.0f", weight);
  // Serial.println("Weight textmessage: " + String(textmessage));
  client.publish(TopicsToPublish[0].c_str(),textmessage);

  sprintf(textmessage, "%.12f", raw);
  // Serial.println("Raw textmessage: " + String(textmessage));
  // Serial.println();
  client.publish(TopicsToPublish[1].c_str(),textmessage);
}

void WiFiEvent(WiFiEvent_t event)
{
  switch (event) {
    case ARDUINO_EVENT_ETH_START:
      Serial.println("ETH Started");
      //set eth hostname here
      ETH.setHostname("esp32-ethernet");
      break;
    case ARDUINO_EVENT_ETH_CONNECTED:
      Serial.println("ETH Connected");
      Serial.println(ETH.macAddress());
      break;
    case ARDUINO_EVENT_ETH_GOT_IP:
      Serial.print("ETH MAC: ");
      Serial.print(ETH.macAddress());
      Serial.print(", IPv4: ");
      Serial.print(ETH.localIP());
      if (ETH.fullDuplex()) {
        Serial.print(", FULL_DUPLEX");
      }
      Serial.print(", ");
      Serial.print(ETH.linkSpeed());
      Serial.println("Mbps");
      eth_connected = true;

      //Configure Eth params
      initialiseMQTT();
      break;
    case ARDUINO_EVENT_ETH_DISCONNECTED:
      Serial.println("ETH Disconnected");
      eth_connected = false;
      break;
    case ARDUINO_EVENT_ETH_STOP:
      Serial.println("ETH Stopped");
      eth_connected = false;
      break;
    default:
      break;
  }
}

void IRAM_ATTR ISR(){
  detachInterrupt(BUTTON1);
  rebounds += 1;
  if (rebounds == 1){
    SPIFFS.format();
    delay(10);
    esp_restart();
    attachInterrupt(BUTTON1, ISR, FALLING);
  }
  else{
    attachInterrupt(BUTTON1, ISR, FALLING);
  }
}

void testClient(const char * host, uint16_t port)
{
  Serial.print("\nconnecting to ");
  Serial.println(host);

  WiFiClient client;
  if (!client.connect(IPAddress(192,168,1,147), port,6000)) {
    Serial.println("connection failed");
    return;
  }
  client.printf("GET / HTTP/1.1\r\nHost: %s\r\n\r\n", host);
  while (client.connected() && !client.available());
  while (client.available()) {
    Serial.write(client.read());
  }

  Serial.println("closing connection\n");
  client.stop();
}

void setup()
{
  Serial.begin(115200);
  Serial.println("Hi there");
  pinMode(BUTTON1, INPUT);
  pinMode(ETH_NRST, OUTPUT);
  pinMode(LED_BLUE, OUTPUT);
  pinMode(ADS_PDWN, OUTPUT);
  digitalWrite(ETH_NRST, HIGH);
  delay(10);
  digitalWrite(ETH_NRST, LOW);
  delay(1);
  digitalWrite(ETH_NRST, HIGH);
  delay(10);
  
  if(SPIFFS.begin()){
    readSPIFFS();
  }

  initialiseDR();
  // fillTopics();

  adc.begin();
  digitalWrite(ADS_SCLK, LOW);
  digitalWrite(ADS_PDWN, HIGH);
  
  // uint8_t mac[] = {0xAC,0x67,0xB2,0x2B,0xA3,0x01};
  // esp_eth_ioctl(ETH.eth_handle, ETH_CMD_S_MAC_ADDR, mac);

  WiFi.onEvent(WiFiEvent);
  ETH.begin();
  delay(5000);

  Serial.println("Server IP: " + DR.ServerIP.toString());
  Serial.println("Server Port: " + String(DR.ServerPort));

  // ETH.config(ETH.localIP(),ETH.gatewayIP(),ETH.subnetMask(),{8,8,8,8},ETH.gatewayIP());
  // delay(2000);
  
  // initialiseMQTT();

  // DR.ServerIP = {192,168,1,134}; //Ezequiel
  // DR.ServerPort = 1883;
  // DR.ServerIP = {5,196,95,208}; //Mosquitto
  // client.setServer(DR.ServerIP, DR.ServerPort);
  client.setServer(DR.ServerIP, DR.ServerPort);
  client.setCallback(callback);
  client.setKeepAlive(8);
  reconnect();

  attachInterrupt(BUTTON1, ISR, FALLING);
}

void loop()
{
  if (!client.connected()) {
    digitalWrite(LED_BLUE, HIGH);
    long now = millis();
    if (now - lastReconnectAttempt > 5000) {    
      // Attempt to reconnect
      reconnect();
      lastReconnectAttempt = millis();
    }
  }
  else{
    digitalWrite(LED_BLUE, LOW);
    // testClient("google.com", 80);
    // delay(1000);
  }

  if(millis() - lastValues >= DR.frequency){
    lastValues = millis();
    sendValue();
    readValue();
  }

  client.loop();
}